@@ -3,22 +3,38 @@ import warnings
 import base64
 import json
 import pickle
+import terragon
 import urllib2
 import urllib
 import types
 import websocket
 import uuid
+import tempfile
 import zlib
 import re
 import os
 import os.path
 import subprocess
+from poster.encode import multipart_encode
+from poster.streaminghttp import register_openers
+from progressbar import ProgressBar, Percentage, Bar, FileTransferSpeed, ETA
 
 from deployment.models import YhatModel
 from deployment.save_session import save_function, _get_source
-from requirements import merge
+from .utils import progressbarify, sizeof_fmt
+
+devnull = open(os.devnull, "w")
+
+# If we can't import everything we need to detect requirements from
+# this version of pip, then we just warn and turn off the feature.
+try:
+    from requirements import getExplicitRequirmets, getImplicitRequirements
+except ImportError:
+    warnings.warn("Unable to use this version of pip. Requirements detection disabled. Consider upgrading pip.")
+    DETECT_REQUIREMENTS = False
+else:
+    DETECT_REQUIREMENTS = True
 
-from .utils import progressbarify
 
 BASE_URI = "http://api.yhathq.com/"
 
@@ -39,7 +55,7 @@ class API(object):
         self.base_uri = base_uri
         self.headers = {'Content-Type': 'application/json'}
 
-    def get(self, endpoint, params):
+    def _get(self, endpoint, params):
         """
         Parameters
         ----------
@@ -66,7 +82,7 @@ class API(object):
         except Exception, e:
             raise e
 
-    def post(self, endpoint, params, data, pb=False):
+    def _post(self, endpoint, params, data, pb=False):
         """
         Parameters
         ----------
@@ -117,34 +133,61 @@ as a pandas DataFrame. If you're still having trouble, please contact:
             raise e
             print "Message: %s" % str(rsp)
 
-    def handshake(self, model_name, model_owner=None):
-        """
-        Parameters
-        ----------
-        model_name: string
-            name of the model you want to connect to
-        model_owner: string
-            username of the model owner for shared models. optional
+    def _post_file(self, endpoint, params, data, pb=True):
+        # Register the streaming http handlers with urllib2
+        register_openers()
 
-        Returns
-        -------
-        ws: WebSocket connection
-            a connection to the model's WebSocketServer
-        """
-        ws_uri = "{BASE_URI}/{USERNAME}/models/{MODEL_NAME}/"
-        ws_base = self.base_uri.replace("http://", "ws://")
-        username = self.username if model_owner is None else model_owner
-        ws_uri = ws_uri.format(BASE_URI=ws_base, USERNAME=username,
-                               MODEL_NAME=model_name)
-        ws_uri = "%s?username=%s&apikey=%s" % (ws_uri, self.username, self.apikey)
-        ws = websocket.create_connection(ws_uri)
-        auth = {
-            "username": self.username,
-            "apikey": self.apikey
+        widgets = ['Transfering Model: ', Bar(), Percentage(), ' ', ETA(), ' ', FileTransferSpeed()]
+        pbar = ProgressBar(widgets=widgets).start()
+        def progress(param, current, total):
+            if not param:
+                return
+            pbar.maxval = total
+            pbar.update(current)
+
+        # headers contains the necessary Content-Type and Content-Length
+        # datagen is a generator object that yields the encoded parameters
+        f = tempfile.NamedTemporaryFile(mode='wb', prefix='tmp_yhat_', delete=False)
+        model_name = data['modelname'] + ".yhat"
+        try:
+            data = json.dumps(data)
+        except UnicodeDecodeError, e:
+            raise Exception("Could not serialize into JSON. String is not utf-8 encoded `%s`" % str(e.args[1]))
+        zlib_compress(data, f)
+        f.close()
+
+        f2 = open(f.name, "rb")
+        datagen, headers = multipart_encode({model_name: f2}, cb=progress)
+
+        url = self.base_uri + endpoint + "?" + urllib.urlencode(params)
+        req = urllib2.Request(url, datagen, headers)
+        auth = '%s:%s' % (params['username'], params['apikey'])
+        base64string = base64.encodestring(auth).replace('\n', '')
+        req.add_header("Authorization", "Basic %s" % base64string)
+        # Actually do the request, and get the response
+        try:
+            response = urllib2.urlopen(req)
+        except urllib2.HTTPError, e:
+            if e.code > 200:
+                responseText = e.read()
+                sys.stderr.write("\nDeployment error: " + responseText)
+                return { "status": "error", "message": responseText }
+            else:
+                sys.stderr.write("\nError in HTTP connection")
+                return { "status": "error", "message": "Error in HTTP connection." }
+        except Exception, e:
+            sys.stderr.write("\nDeployment error: " + str(e))
+            return { "status": "error", "message": str(e) }
+        rsp = response.read()
+        pbar.finish()
+        # clean up after we're done
+	f2.close()
+	os.unlink(f.name)
+        reply = {
+            "status": "OK",
+            "message": "Model successfully uploaded. Your model will begin building momentarily. Please see %s for more details" % self.base_uri
         }
-        ws.send(json.dumps(auth))
-        return ws
-
+        return reply
 
 class Yhat(API):
 
@@ -171,8 +214,9 @@ class Yhat(API):
         self.headers = {'Content-Type': 'application/json'}
         self.q = {"username": self.username, "apikey": apikey}
         if self.base_uri != BASE_URI:
-            if self._authenticate() is False:
-                raise Exception("Incorrect username/apikey!")
+            e = self._authenticate()
+            if e is not None:
+                raise Exception("Failed to authenticate: %s" % e)
 
     def _check_obj_size(self, obj):
         """
@@ -200,16 +244,15 @@ class Yhat(API):
         """
         Returns
         -------
-        authed: boolean
+        error: None or Exception
             verifies your API credentials are valid
         """
-        authed = True
         try:
-            response = self.post('verify', self.q, {})
+            response = self._post('verify', self.q, {})
             error = response["success"]
+            return None
         except Exception, e:
-            authed = False
-        return authed
+            return e
 
     def _convert_to_json(self, data):
         """
@@ -255,7 +298,7 @@ class Yhat(API):
         model_owner: string
             username of the model owner for shared models. optional
         raw_input: bool
-            whether or not to attempt to parse the incoming JSON. for calling R 
+            whether or not to attempt to parse the incoming JSON. for calling R
             models only
 
         Returns
@@ -274,65 +317,9 @@ class Yhat(API):
         else:
             data = {"data": data}
             endpoint = 'predict'
-        return self.post(endpoint, q, data)
+        return self._post(endpoint, q, data)
 
-    def predict_ws(self, data):
-        """
-
-        Parameters
-        ----------
-        data: dictionary or data frame
-            data required to make a single prediction. this can be a dict or
-            a dataframe
-
-        Returns
-        -------
-        id: string
-            the id of the event; this will come in handy when it
-            is receieved by the WebSocket client (remember, this is
-            asynchronous)
-        """
-        if self.ws is None:
-            msg = """In order to make predictions with WebScokets, you
-need to connect to the server first. try running "connect_to_socket"
-"""
-            raise Exception(msg)
-        data = self._convert_to_json(data)
-        data['_id'] = str(uuid.uuid4())
-        self.ws.send(json.dumps(data))
-        return data['_id']
-
-    def yield_results(self):
-        """
-        Listens to the WebSocket and awaits completed predictions
-
-        Returns
-        -------
-        iterator: generator
-            yields a new prediction
-        """
-        if self.ws is None:
-            msg = """In order to make predictions with WebScokets, you
-need to connect to the server first. try running "connect_to_socket"
-"""
-            raise Exception(msg)
-
-        while True:
-            yield self.ws.recv()
-
-    def connect_to_socket(self, model, model_owner=None):
-        """
-        Connects to the model's WebSocket endpoint. This is a pre-requisite for
-        making predictions via the WebSocketServer.
-
-        model: string
-            name of the model you want to connect to
-        model_owner: string
-            username of the model owner for shared models. optional
-        """
-        self.ws = self.handshake(model, model_owner)
-
-    def _extract_model(self, name, model, session):
+    def _extract_model(self, name, model, session, autodetect, verbose=0):
         """
         Extracts source code and any objects required to deploy the model.
 
@@ -344,26 +331,62 @@ need to connect to the server first. try running "connect_to_socket"
             an instance of a Yhat model
         session: globals()
             your Python's session variables (i.e. "globals()")
+        verbose: int
+            log level
         """
+        code = ""
+        print "extracting model"
+
         if 1 == 2 and _get_source(YhatModel.execute) == _get_source(model.execute):
             msg = """'execute' method was not implemented.
             If you believe that you did implement the 'execute' method,
             check to make sure there isn't an indentation error."""
             raise Exception(msg)
 
-        bundle = save_function(model, session)
+        bundle = save_function(model, session, verbose=verbose)
         bundle["largefile"] = True
         bundle["username"] = self.username
         bundle["language"] = "python"
         bundle["modelname"] = name
         bundle["className"] = model.__name__
-        bundle["reqs"] = "\n".join(
-            str(r) for r in merge(session, getattr(model, "REQUIREMENTS", ""))
-        )
+        bundle["code"] = code + "\n" + bundle.get("code", "")
+
+        # REQUIREMENTS
+        if DETECT_REQUIREMENTS and autodetect:
+            requirements = getImplicitRequirements(model, session)
+        else:
+            requirements = getExplicitRequirmets(model, session)
+        bundle["reqs"] = requirements
+
+        # MODULES
+        modules = bundle.get("modules", [])
+        if modules:
+            print "model source files"
+            for module in modules:
+                name = module["name"]
+                parent_dir = module.get("parent_dir", "")
+                if parent_dir != "":
+                    name = os.path.join(parent_dir, name)
+                print " [+]", name
+
+        # OBJETCS
+        objects = bundle.get("objects", {})
+        if objects:
+            print "model variables"
+            for name, pkl in objects.iteritems():
+                try:
+                    obj = terragon.loads_from_base64(pkl)
+                    t = type(obj)
+                    del obj
+                except Exception as e:
+                    print "ERROR pickling object:", name
+                    raise e
+                size = 3. * float(len(pkl)) / 4.
+                print " [+]", name, t, sizeof_fmt(size)
 
         return bundle
 
-    def deploy(self, name, model, session, sure=False, packages=[], patch=None):
+    def deploy(self, name, model, session, sure=False, packages=[], patch=None, dry_run=False, verbose=0, autodetect=True):
         """
         Deploys your model to a Yhat server
 
@@ -379,6 +402,9 @@ need to connect to the server first. try running "connect_to_socket"
             if true, then this will force a deployment (like -y in apt-get).
             if false or blank, this will ask you if you're sure you want to
             deploy
+        autodetect: flag for using the requirement auto-detection feature.
+            if False, you should explicitly state the packages required for
+            your model, or it may not run on the server.
         """
         # first let's check and make sure the user actually wants to deploy
         # a new version
@@ -387,91 +413,34 @@ need to connect to the server first. try running "connect_to_socket"
         if not isinstance(packages, list):
             raise Exception(
                 "`packages` must be a list of ubuntu packages to install")
-        if sure is False:
+        if (not sure) and (not dry_run):
             sure = raw_input("Are you sure you want to deploy? (y/N): ")
             if sure.lower() != "y":
                 print "Deployment canceled"
                 sys.exit()
-        bundle = self._extract_model(name, model, session)
+        bundle = self._extract_model(name, model, session, verbose=verbose, autodetect=autodetect)
         bundle['packages'] = packages
         if isinstance(patch, str)==True:
             patch = "\n".join([line.strip() for line in patch.strip().split('\n')])
             bundle['code'] = patch + "\n" + bundle['code']
+
+        if dry_run:
+            return {"status": "ok", "info": "dry run complete"}
         if self._check_obj_size(bundle) is False:
             # we're not going to deploy; model is too big, but let's give the
             # user the option to upload it manually
-            print "Model is to large to deploy over HTTP"
-            should_we_deploy = raw_input(
-                "Would you like to upload manually? (Y/n): ")
-            if should_we_deploy.lower() == "y" or should_we_deploy == "":
-                self.deploy_to_file(name, model, session)
+            raise Exception("Model is too large to deploy over HTTP")
         else:
             # upload the model to the server
-            print "Uploading model data"
-            data = self.post("deployer/model", self.q, bundle, pb=True)
-            print "Model uploaded"
+            data = self._post_file("deployer/model/large", self.q, bundle, pb=True)
             return data
 
-    def deploy_to_file(self, name, model, session, compress=True, packages=[], patch=None):
-        """
-        Bundles a local version of your model that can be manually uploaded to
-        the server.
 
-        Parameters
-        ----------
-        name: string
-            name of your model
-        model: YhatModel
-            an instance of a Yhat model
-        session: globals()
-            your Python's session variables (i.e. "globals()")
-        """
-        if not re.match("^[A-Za-z0-9_]+$", name):
-            raise Exception("Model name must only contain: [A-Za-z0-9_]")
-        if not isinstance(packages, list):
-            raise Exception(
-                "`packages` must be a list of ubuntu packages to install")
-        bundle = self._extract_model(name, model, session)
-        bundle['apikey'] = self.apikey
-        bundle['packages'] = packages
-        if isinstance(patch, str)==True:
-            patch = "\n".join([line.strip() for line in patch.strip().split('\n')])
-            bundle['code'] = patch + "\n" + bundle['code']
-        filename = "%s.yhat" % name
-        with open(filename, "w") as f:
-            bundle = json.dumps(bundle)
-            if compress is True:
-                bundle = zlib.compress(bundle)
-            f.write(bundle)
+def zlib_compress(data, to):
+    step = 4 << 20 # 4MiB
+    c = zlib.compressobj()
 
-        print "Model successfully bundled to file:"
-        print "\t%s/%s.yhat" % (os.getcwd(), name)
-        msg = "To deploy, visit %s and upload %s."
-        upload_url = os.path.join(self.base_uri, "model", "upload")
-        msg = msg % (upload_url, "%s.yhat" % name)
-        print msg
-        return filename
+    for i in xrange(0, len(data), step):
+        to.write(c.compress(data[i:i+step]))
 
-    def deploy_with_scp(self, name, model, sessions,
-                        compress=True, packages=[], pem_path=None):
-        if pem_path is None:
-            raise Exception("Please specify your pem file for "
-                            "authentication through the `pem_path` argument")
-            return
-        if not os.path.isfile(pem_path):
-            raise Exception("No file found under '%s'" % pem_path)
-        print "Deploying to file"
-        filename = self.deploy_to_file(
-            name, model, sessions, compress=compress, packages=packages)
-        print "Sending over scp"
-        http_re = re.compile("^http://")
-        server_uri = http_re.sub("", self.base_uri.strip())
-        server_uri = server_uri.strip("/")
-        scp_cmd = "scp -i %s %s ubuntu@%s:~/" % (
-            pem_path, filename, server_uri)
-        subprocess.check_call(scp_cmd, shell=True)
-        ssh_cmd = """ssh -i %s ubuntu@%s
-        'sudo mv ~/%s
-        /var/yhat/headquarters/uploads/'""" % (pem_path, server_uri, filename)
-        subprocess.check_call(ssh_cmd, shell=True)
-        os.remove(filename)
+    to.write(c.flush())
